"""
    Create and run more than 1000 that work together. They share a list, input_list,
    and an integer variable.
    When they run, they take one random item from the input list and add it to that variable
    When they finish, the input list should be empty.

    Requirements:
        * the length of the input list is equal to the number of threads
        * the threads have a random delay when running (see time.sleep)
        * before starting the threads compute the sum of the input list
        * after the threads stopped, the shared variable should be identical to the sum of the
        input list

    Hint:
        * in CPython some operations on data structures are atomic, some are not. Use locks when
        the operations are not thread-safe.
        useful links:
        https://docs.python.org/2/faq/library.html#what-kinds-of-global-value-mutation-are-thread-safe
"""

import time
import random
import threading
import sys

shared_var = 0


class MyThread(threading.Thread):
   
    def __init__(self, input_list, list_lock, var_lock):
        threading.Thread.__init__(self)
        self.input_list = input_list
        self.shared_var = shared_var
        self.list_lock = list_lock
        self.var_lock = var_lock
        
    def run(self):
        time.sleep(random.randint(0, 3))

        # we need a lock even though pop is atomic, because we also call the len method. Between
        # the call of len and pop we don't want another thread to modify the list. Also,
        # it is not well documented if len is atomic.

        with self.list_lock:
            elem_index = random.randint(0, len(input_list) - 1)
            elem = input_list.pop(elem_index)
        
        with var_lock:
            global shared_var
            shared_var += elem


def usage():
    print "Usage: %s num_threads" % sys.argv[0]


if __name__ == "__main__":

    if len(sys.argv) != 2:
        usage()
        sys.exit(1)

    num_threads = int(sys.argv[1])

    threads = []

    input_list = [random.randint(0, 500) for i in range(num_threads)]
    initial_sum = reduce(lambda x, y: x+y, input_list)
    print " ".join([str(x) for x in input_list])

    var_lock = threading.Lock()
    list_lock = threading.Lock()

    shared_var = 0
    
    for i in range(num_threads):
        threads.append(MyThread(input_list, list_lock, var_lock))
        threads[-1].start()

    for t in threads:
        t.join()

    print initial_sum
    print shared_var
